from .lzhw_alg import LZHW
from pickle import dump, load, HIGHEST_PROTOCOL
from .compress_util import _reader
import pandas as pd
from lzw_c import *
from operator import itemgetter
from .para_util import lzhw_para, para_decompress

class CompressedDF:
    def __init__(self, df, selected_cols="all"):
        if selected_cols == "all":
            selected = range(df.shape[1])
        else:
            selected = selected_cols
        self.columns = list(itemgetter(*selected)(df.columns))
        self.compressed = lzhw_para(df, selected)

    def save_to_file(self, file):
        with open(file, "wb") as output:
            cols = [i.replace(" ", "__") for i in self.columns]
            dump(lzw_compress(" ".join(cols)), output, HIGHEST_PROTOCOL)
            for i in range(len(self.columns)):
                dump(self.compressed[i].compressed, output, HIGHEST_PROTOCOL)
                dump(self.compressed[i].sequences, output, HIGHEST_PROTOCOL)


def decompress_df_from_file(file, selected_cols="all", n_rows=0):
    with open(file, "rb") as input:
        cols = load(input)
        cols = lzw_decompress(cols).split()
        cols = [i.replace("__", " ") for i in cols]
        if selected_cols == "all":
            selected = range(len(cols))
        else:
            if any(isinstance(c, str) for c in selected_cols):
                selected_cols = [i for i, v in enumerate(cols) if v in selected_cols]
            selected = selected_cols
        sequences, triplets = _reader(input, len(cols), selected)
        decompressed = para_decompress(sequences, triplets, n_rows)
        names = list(itemgetter(*selected)(cols))
        df = dict(zip(names, decompressed))

    if n_rows > 0:
        df = {k: v[:n_rows] for k, v in df.items()}
    df = pd.DataFrame(df)
    return df




#from lzw_c import *
from cpython cimport int as Integer

ctypedef fused str_num:
    Integer
    int
    str
    double
    float

cdef list lz77compress(list data):
    cdef int l = len(data)
    cdef int _sliding_window = 512
    cdef int current_location = 0
    cdef list triplets = []
    cdef tuple triplet
    cdef int ol
    cdef int ols
    while current_location < l:
        triplet = triplet_encode(data, current_location, _sliding_window)
        triplets.append(triplet)
        if not triplet[1]:
            ols = 0
        else:
            ols = triplet[1]
        current_location += 1 + ols
    return triplets

cdef tuple triplet_encode(list data, int current_location, int sliding_window):
    cdef int _match_len = 0
    cdef int match_offset = 0
    cdef int buffer_start = 1
    cdef int buffer_slide = current_location - buffer_start
    cdef int matchlen
    while buffer_slide >= 0 and buffer_start < sliding_window:
        matchlen = match(data, current_location, buffer_slide)
        if matchlen > _match_len:
            _match_len = matchlen
            match_offset = buffer_start
        buffer_start += 1
        buffer_slide -= 1
    if match_offset == 0:
        mo = None
        _ml = None
    else:
        mo = match_offset
        _ml = _match_len
    literal = data[current_location + _match_len]
    #cdef Integer literal = lzw_compress(liter)
    return mo, _ml, literal

cdef int match(list data, int current_location, int buffer_slide):
    cdef int matchlen = 0
    while current_location + matchlen + 1 < len(data):
        if data[current_location + matchlen] != data[buffer_slide + matchlen]:
            break
        matchlen += 1
    return matchlen

cpdef list lz77_compress(list data):
    cdef list triplets = lz77compress(data)
    return triplets

cdef list lz77decompress(list compressed, int n_rows):
    cdef tuple triplet
    cdef str decomp
    cdef list decompressed = []
    cdef int l #= len(decompressed)
    cdef int offset #= int(triplet[0])
    cdef int length #= int(triplet[1)
    #cdef Integer liter #= triplet[2]
    #cdef str literal
    cdef int current_location #= l - offset
    cdef int i
    for triplet in compressed:
        if not triplet[0]:
            offset = 0
            length = 0
        else:
            offset = triplet[0]
            length = triplet[1]
        literal = triplet[2]
        #literal = lzw_decompress(liter)
        if offset > 0:
            l = len(decompressed)
            current_location = l - offset
            for i in range(length):
                decomp = decompressed[current_location]
                decompressed.append(decomp)
                current_location += 1
        decompressed.append(literal)
        if n_rows > 0 and len(decompressed) >= n_rows:
            break
    return decompressed

cpdef list lz77_decompress(list compressed, int n_rows = 0):
    cdef list decompressed = lz77decompress(compressed, n_rows)
    return decompressed



from joblib import Parallel, delayed
from .lzhw_alg import LZHW
from tqdm import tqdm
from .compress_util import lzhw_decompress


def lzhw_para(df, selected):
    compressed = Parallel(n_jobs=-2, max_nbytes=None, backend="multiprocessing")(delayed(
        LZHW)(list(df.iloc[:, i])) for i in tqdm(selected))
    return compressed

def para_decompress(sequences, triplets, n_rows):
    decompressed = Parallel(n_jobs=-2, max_nbytes=None, backend="multiprocessing")(delayed(
        lzhw_decompress)(sequences[i], triplets[i], n_rows) for i in tqdm(range(len(sequences))))
    return decompressed



from lzw_c import *
from lz77c import lz77_decompress
from pickle import load

def glue_seq(seq, last_separate = False):
    if last_separate:
        s = seq.split()
        return " ".join(s[:-1]), s[-1]
    else:
        return " ".join(seq)

def code_filling(huff_codes, n):
    sequences = {}
    codes = {}
    for seq, code in huff_codes.items():
        cd = int("1" + code, 2)
        if n == "literal_str":
            sequences[cd] = lzw_compress(str(seq))
        elif n == "literal":
            sequences[cd] = eval(seq) if seq != "nan" else seq
        else:
            sequences[cd] = seq
        codes[seq] = code
    return sequences, codes

def huffman_decode(sequences, compressed, n):
    bitstring = bin(compressed)[2:]
    bit = ""
    for b in bitstring[1:]:  # after preceding 1
        bit += b
        bit_int = int("1" + bit, 2)
        if bit_int in sequences:
            if n == "literal_str":
                org = lzw_decompress(sequences[bit_int])
            else:
                org = str(sequences[bit_int])
            yield org
            bit = ""

def org_shaping(seq, bits, n):
    org = " ".join(huffman_decode(seq, bits, n)).split()
    org = [i.replace("__", " ") for i in org]
    if "None" in org:
        org = [eval(i) for i in org]
    return org

def lz77_decode(triplets, n_rows):
    triplets = list(zip(triplets[0], triplets[1], triplets[2]))
    decomp = lz77_decompress(triplets, n_rows)
    return decomp

def lzhw_decompress(sequences, triplets, n_rows):
    if "lz77" in sequences:
        decomp = lz77_decode(triplets, n_rows)
    else:
        trplts = []
        for n, i in zip(sequences.keys(), range(len(triplets))):
            triplet = org_shaping(sequences[n], triplets[i], n)
            trplts.append(triplet)
        decomp = lz77_decode(trplts, n_rows)
    return decomp

def _reader(input, cols_len, selected):
    triplets = []
    sequences = []
    for i in range(cols_len):
        triplet = load(input)
        sequence = load(input)
        if i not in selected:
            continue
        else:
            triplets.append(triplet)
            sequences.append(sequence)
    return sequences, triplets


#!/usr/bin/env python

def main():
    import lzhw
    import pandas as pd
    import argparse
    import os
    from subprocess import call
    from time import time

    ## This script and the solution to convert xlsx into csv was thanks to the answer found here:
    ## https://stackoverflow.com/questions/28766133/faster-way-to-read-excel-files-to-pandas-dataframe
    ## and here: https://stackoverflow.com/questions/1858195/convert-xls-to-csv-on-command-line
    vbscript = """if WScript.Arguments.Count < 3 Then
        WScript.Echo "Please specify the source and the destination files. Usage: ExcelToCsv <xls/xlsx source file> <csv destination file> <worksheet number (starts at 1)>"
        Wscript.Quit
    End If

    csv_format = 6

    Set objFSO = CreateObject("Scripting.FileSystemObject")

    src_file = objFSO.GetAbsolutePathName(Wscript.Arguments.Item(0))
    dest_file = objFSO.GetAbsolutePathName(WScript.Arguments.Item(1))
    worksheet_number = CInt(WScript.Arguments.Item(2))

    Dim oExcel
    Set oExcel = CreateObject("Excel.Application")

    Dim oBook
    Set oBook = oExcel.Workbooks.Open(src_file)
    oBook.Worksheets(worksheet_number).Activate

    oBook.SaveAs dest_file, csv_format

    oBook.Close False
    oExcel.Quit
    """

    def is_number(s):
        try:
            float(s)
            return True
        except ValueError:
            return False

    def csv_reader(file, cols, col_arg, nh_arg):
        if nh_arg:
            h = None
        else:
            h = 0
        if col_arg:
            cols_used = cols.split(",")
            if is_number(cols_used[0]):
                cols_used = [int(i) - 1 for i in cols_used]
        else:
            cols_used = None

        data = pd.read_csv(file, header=h, usecols=cols_used)
        data.columns = list(map(str, data.columns))
        return data

    parser = argparse.ArgumentParser(
        description="LZHW is a tabular data compression tool. It is used to compress excel, csv and any flat file. Version: 0.0.9")
    parser.add_argument("-d", "--decompress", help="decompress input into output",
                        action="store_true", default=False)
    parser.add_argument("-f", "--input", help="input file to be (de)compressed",
                        type=str, required=True)
    parser.add_argument("-o", "--output", help="output where to save result",
                        type=str, required=True)
    parser.add_argument("-c", "--columns", nargs="+",
                        help="select specific columns by names or indices (1-based) to compress or decompress",
                        type=str,
                        required=False)
    parser.add_argument("-r", "--rows",
                        help="select specific rows to decompress (1-based)", type=str,
                        required=False)
    parser.add_argument("-nh", "--no-header", help="skip header / data to be compressed has no header",
                        action="store_true", default=False)
    args = vars(parser.parse_args())

    file = args["input"]
    output = args["output"]

    if args["columns"]:
        cols = args["columns"][0]
    else:
        cols = "all"

    if args["rows"]:
        n_rows = int(args["rows"])
    else:
        n_rows = 0

    if args["decompress"]:
        start = time()
        if cols != "all":
            cols = cols.split(",")
            if is_number(cols[0]):
                cols = [int(i) - 1 for i in cols]

        decompressed = lzhw.decompress_df_from_file(file, cols, n_rows)
        decompressed.fillna("", inplace=True)
        decompressed = decompressed.replace("nan", "", regex=True)
        if "xls" in output:
            # decompressed.reset_index(drop = True, inplace = True)
            options = {}
            options["strings_to_formulas"] = False
            options["strings_to_urls"] = False
            writer = pd.ExcelWriter(output, engine="xlsxwriter", options=options)
            decompressed.to_excel(writer, output.split(".xls")[0], index=False)
            writer.save()
            # decompressed.to_excel(output, index=False, encoding = "utf8")
        if "csv" in output:
            decompressed.to_csv(output, index=False)
        else:
            with open(output, "w") as o:
                decompressed.to_string(o, index=False)
        print("Finalizing Decompression ...")
        print(f"Creating {output} file ...")
        print("time taken: ", (time() - start) / 60, " minutes")
        print("Decompressed Successfully")

    else:
        start = time()
        if "xls" in file:
            print("Reading files, Can take 1 minute or something ...",
                  "\nRunning CScript.exe to convert xls file to csv for better performance", "\n")
            f = open("excel_to_csv.vbs", "w")
            f.write(vbscript)
            f.close()
            csv_file = file.split(".xls")[0] + "1" + ".csv"
            call(["cscript.exe", "excel_to_csv.vbs", file, csv_file, "1"])
            os.remove("excel_to_csv.vbs")

            data = csv_reader(csv_file, cols, args["columns"], args["no_header"])

            os.remove(csv_file)

        elif "csv" in file:
            print("Reading files ...")
            data = csv_reader(file, cols, args["columns"], args["no_header"])

        else:
            with open(file, "r") as i:
                data = i.read()

        comp_df = lzhw.CompressedDF(data)
        print("Finalizing Compression ...")
        comp_df.save_to_file(output)
        print(f"Creating {output} file ...")
        print("time taken: ", (time() - start) / 60, " minutes")
        print("Compressed Successfully")


if __name__ == "__main__":
    import multiprocessing

    multiprocessing.freeze_support()
    multiprocessing.Process(target=main).start()
